import 'aggregation.dart';
import 'quantity.dart';

class TimeSeries<N extends num> {
  final DateTime start;
  final TimeUnit timeUnit;
  final Unit unit;
  final List<N> values;

  TimeSeries({
    this.start,
    this.timeUnit,
    this.unit,
    this.values,
  });

  TimeSeries.fromJson(Map<String, dynamic> json)
      : this(
          start: DateTime.parse(json['start']),
          timeUnit: TimeUnit.parse(json['timeUnit']),
          unit: Unit.fromJson(json['unit']),
          values: json['values'].map<N>((value) => _valueFromJson(value) as N).toList(),
        );

  Map<DateTime, Quantity<N>> get points {
    var series = <DateTime, Quantity<N>>{};
    var currentTimestamp = start;
    values.forEach((value) {
      series[currentTimestamp] = Quantity(unit, value);
      currentTimestamp = timeUnit.addTo(currentTimestamp);
    });
    return series;
  }

  int get length => values.length;

  TimeSeries<O> aggregate<O extends num>(TimeUnit timeUnit, Aggegator<N, O> aggregator) {
    if (timeUnit == null) {
      throw ArgumentError('Time unit cannot be null');
    }
    var segmented = _segmentValues(timeUnit);
    var aggregated = segmented.map((bucket) => aggregator(bucket)).toList();
    return TimeSeries(start: start, timeUnit: timeUnit, values: aggregated, unit: unit);
  }

  List<TimeSeries<N>> segment(TimeUnit timeUnit) {
    var currentTimestamp = start;
    return _segmentValues(timeUnit).map((values) {
      var timeSeries = TimeSeries(start: currentTimestamp, timeUnit: this.timeUnit, unit: unit, values: values);
      currentTimestamp = timeUnit.addTo(currentTimestamp);
      return timeSeries;
    }).toList();
  }

  List<List<N>> _segmentValues(TimeUnit timeUnit) {
    var result = <List<N>>[];
    var currentValues = <N>[];
    var index = 0;
    var currentTimeStamp = start;
    var segment = timeUnit.extract(currentTimeStamp);
    DateTime nextSegment;
    while (index < values.length) {
      do {
        currentValues.add(values[index]);
        index++;
        currentTimeStamp = this.timeUnit.addTo(currentTimeStamp);
        nextSegment = timeUnit.extract(currentTimeStamp);
      } while (nextSegment == segment && index < values.length);
      segment = nextSegment;
      result.add(currentValues.toList());
      currentValues.clear();
    }
    return result;
  }

  Map<String, dynamic> toJson() => {
        'start': start.toIso8601String(),
        'timeUnit': timeUnit.name,
        'unit': unit.toJson(),
        'values': values.map(_valueToJson).toList(),
      };

  static dynamic _valueFromJson(dynamic value) => value is num ? value : DataPoint.fromJson(value);

  static dynamic _valueToJson(dynamic value) => value is num ? value : value.toJson();

  TimeSeries<N> merge(TimeSeries<N> other) {
    if (timeUnit != other.timeUnit) {
      throw ArgumentError('Cannot merge $this with $other, time unit differs');
    }
    if (unit != other.unit) {
      throw ArgumentError('Cannot merge $this with $other, unit differs');
    }
    var earliestStart = start.isBefore(other.start) ? start : other.start;
    var currentTimeStamp = earliestStart;
    var newValues = <N>[];
    while (currentTimeStamp.isBefore(timeUnit.addTo(start, values.length)) ||
        currentTimeStamp.isBefore(timeUnit.addTo(other.start, other.values.length))) {
      newValues.add((getValueAt(currentTimeStamp) + other.getValueAt(currentTimeStamp)) as N);
      currentTimeStamp = timeUnit.addTo(currentTimeStamp);
    }
    return TimeSeries(start: earliestStart, timeUnit: timeUnit, unit: unit, values: newValues);
  }

  @override
  String toString() {
    return 'TimeSeries{start: $start, timeUnit: $timeUnit, unit: $unit}';
  }

  N getValueAt(DateTime timeStamp) {
    if (timeStamp.isBefore(start) || timeStamp.isAfter(timeUnit.addTo(start, values.length - 1))) {
      return 0 as N;
    }
    var currentTimestamp = start;
    var index = 0;
    while (currentTimestamp.isBefore(timeStamp)) {
      currentTimestamp = timeUnit.addTo(currentTimestamp);
      index++;
    }
    return values[index];
  }
}

class TimeUnit {
  static final TimeUnit second = TimeUnit(
    'second',
    (dt, m) => dt.add(Duration(seconds: m)),
    (dt) => DateTime(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second, 0, 0),
  );
  static final TimeUnit minute = TimeUnit(
    'minute',
    (dt, m) => dt.add(Duration(minutes: m)),
    (dt) => DateTime(dt.year, dt.month, dt.day, dt.hour, dt.minute, 0, 0, 0),
  );
  static final TimeUnit hour = TimeUnit(
    'hour',
    (dt, m) => dt.add(Duration(hours: m)),
    (dt) => DateTime(dt.year, dt.month, dt.day, dt.hour, 0, 0, 0, 0),
  );
  static final TimeUnit day = TimeUnit(
    'day',
    (dt, m) => dt.add(Duration(days: m)),
    (dt) => DateTime(dt.year, dt.month, dt.day, 0, 0, 0, 0, 0),
  );
  static final TimeUnit month = TimeUnit(
    'month',
    (dt, m) => dt.addMonths(m),
    (dt) => DateTime(dt.year, dt.month, 0, 0, 0, 0, 0, 0),
  );
  static final TimeUnit quarter = TimeUnit(
    'quarter',
    (dt, m) => dt.addMonths(3 * m),
    (dt) => DateTime(dt.year, (dt.month / 3).ceil(), 0, 0, 0, 0, 0, 0),
  );
  static final TimeUnit year = TimeUnit(
    'year',
    (dt, m) => dt.addYears(m),
    (dt) => DateTime(dt.year, 0, 0, 0, 0, 0, 0, 0),
  );
  static final TimeUnit decade = TimeUnit(
    'decade',
    (dateTime, multiplicand) => dateTime.addYears(10 * multiplicand),
    (dt) => DateTime(((dt.year) + 1 / 10).ceil(), 0, 0, 0, 0, 0, 0, 0),
  );
  static final TimeUnit century = TimeUnit(
    'century',
    (dateTime, multiplicand) => dateTime.addYears(100 * multiplicand),
    (dt) => DateTime(((dt.year) + 1 / 100).ceil(), 0, 0, 0, 0, 0, 0, 0),
  );
  static final TimeUnit millennium = TimeUnit(
    'millennium',
    (dateTime, multiplicand) => dateTime.addYears(1000 * multiplicand),
    (dt) => DateTime(((dt.year) + 1 / 1000).ceil(), 0, 0, 0, 0, 0, 0, 0),
  );
  static final List<TimeUnit> values = [
    second,
    minute,
    hour,
    day,
    month,
    quarter,
    year,
    decade,
    century,
    millennium,
  ];

  final String name;
  final DateTime Function(DateTime, int) _incrementor;
  final DateTime Function(DateTime) _extractor;

  TimeUnit(this.name, this._incrementor, this._extractor);

  DateTime addTo(DateTime dateTime, [int multiplicand = 1]) => _incrementor(dateTime, multiplicand);

  DateTime extract(DateTime dateTime) => _extractor(dateTime);

  factory TimeUnit.parse(String name) => values.where((element) => element.name == name).first;
}

extension DateTimeAddMonth on DateTime {
  DateTime addMonths(int months) {
    var month = this.month + months;
    var year = this.year;
    while (month > 12) {
      month -= 12;
      year += 1;
    }
    return DateTime(year, month, day, hour, minute, second, millisecond, microsecond);
  }

  DateTime addYears(int years) => DateTime(year + years, month, day, hour, minute, second, millisecond, microsecond);
}
